#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "DEFAULT_PORT",
    "DEFAULT_DB",
    "LIBNAME_MAX_LEN",
    "DATA_SET_MAX_LEN",
    "DEFAULT_HANDLER",
    "DEFAULT_WS_USER",
    "DEFAULT_WS_PASSWORD",
    "DEFAULT_WS_URL",
    "WORK_SCHEMA",
    "DUMMY_FROM",
]

# Default SAS port
DEFAULT_PORT = 8591
# Default database name
DEFAULT_DB = "sas"
# Max library name length
LIBNAME_MAX_LEN = 8
# Max data set name length
DATA_SET_MAX_LEN = 32
# Handle
DEFAULT_HANDLER = "ws"  # "saspy"
# Default web service handler username
DEFAULT_WS_USER = "sasjdbc"
# Default web service handler password
DEFAULT_WS_PASSWORD = "sasjdbc"
# Default web service handler url
DEFAULT_WS_URL = "http://localhost:8099"
# Work schema
WORK_SCHEMA = "work"
# Dummy from when from table is missing
DUMMY_FROM = """
from sashelp.vdctnry
where memname='DICTIONARIES' and name='NAME'
"""



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from pathlib import Path
from typing import Optional

import agate

from dbt.adapters.sas import sas_log, sas_macros
from dbt.adapters.sas.credentials import SasCredentials
from dbt.exceptions import RuntimeException

__all__ = ["AbstractConnectionHandler"]


class AbstractConnectionHandler(abc.ABC):
    def __init__(self, credentials: SasCredentials) -> None:
        self.credentials = credentials

    @abc.abstractmethod
    def submit(self, code: str, note: Optional[str] = None) -> str:
        """Submit code to the SAS server"""
        return NotImplemented

    @abc.abstractmethod
    def select(self, sql: str) -> agate.Table:
        """Execute a SQL select on the SAS server"""
        return NotImplemented

    @abc.abstractmethod
    def endsas(self) -> None:
        """Terminate the SAS session, shutting down the SAS process"""
        return NotImplemented

    def upload_file(self, local_filename: str, remote_filename: str) -> None:
        """Upload a file to the SAS server"""
        data = Path(local_filename).read_text().rstrip("\n")
        code = sas_macros.UPLOAD_FILE.format(remote_filename=remote_filename, data=data)
        self.submit(code, note="sas")

    def delete_file(self, remote_filename: str) -> None:
        """Delete a file from the SAS server"""
        sas_log.note(f"Delete file - Filename={remote_filename}")
        code = sas_macros.DELETE_FILE.format(remote_filename=remote_filename)
        self.submit(code)

    def check_error(self, output: str, ignore_warnings: bool = False):
        """Check for error message in the result log"""
        log_lines = output.splitlines()
        if self.credentials.fail_on_warnings and not ignore_warnings:
            error_lines = [line for line in log_lines if line.startswith("ERROR") or line.startswith("WARNING")]
        else:
            error_lines = [line for line in log_lines if line.startswith("ERROR")]
        if error_lines:
            sas_log.error(output)
            raise RuntimeException(error_lines[0])

    @classmethod
    def load_autoexec(self, credentials: SasCredentials) -> str:
        # Load autoexec
        if credentials.autoexec:
            autoexec = Path(credentials.autoexec).read_text()
        else:
            autoexec = ""
        # Append auto assign libname to autoexec
        if credentials.lib_base_path:
            autoexec += sas_macros.ASSIGN_LIBNAMES.format(path=credentials.lib_base_path)
        return autoexec



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Any, List, Optional, Tuple

import agate

from dbt.adapters.base import BaseAdapter, available
from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.sas import SasConnectionManager, sas_log, sas_macros
from dbt.adapters.sas.column import TEXT, SasColumn
from dbt.adapters.sas.constants import DEFAULT_DB
from dbt.adapters.sas.relation import SasRelation
from dbt.adapters.sas.utils import (
    agate_table_get_column_values,
    get_temp_filename,
    get_temp_local_filename,
    path_join,
)
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.relation import RelationType
from dbt.exceptions import RuntimeException

__all__ = ["SasAdapter"]


class SasAdapter(BaseAdapter):
    ConnectionManager = SasConnectionManager
    Relation = SasRelation
    Column = SasColumn

    """
    Methods:
        - exception_handler
        - date_function
        - list_schemas [OK]
        - drop_relation [OK]
        - truncate_relation [OK]
        - rename_relation [OK]
        - get_columns_in_relation [OK] - TODO: check data types
        - expand_column_types
        - list_relations_without_caching [OK]
        - is_cancelable [OK]
        - create_schema [OK]
        - drop_schema [OK]
        - quote [OK]
        - convert_text_type
        - convert_number_type
        - convert_boolean_type [OK]
        - convert_datetime_type
        - convert_date_type
        - convert_time_type
        - standardize_grants_dict

    Macros:
        - get_catalog
    """

    @classmethod
    def date_function(cls):
        """
        Returns canonical date func
        """
        return "datetime()"  # format=datetime22.

    def debug_query(self) -> None:
        sas_log.note("Debug query")
        connection = self.connections.get_thread_connection()
        connection.handle.submit(sas_macros.NOOP)

    @classmethod
    def is_cancelable(cls) -> bool:
        return False

    def list_schemas(self, database: str) -> List[str]:
        """Get a list of existing libraries"""
        table = self.execute_macro("list_schemas", kwargs={"database": ""})
        return agate_table_get_column_values(table, "libname")

    def list_relations_without_caching(self, schema_relation: BaseRelation) -> List[BaseRelation]:
        """List data sets in the given library, bypassing the cache."""
        table = self.execute_macro("list_relations_without_caching", kwargs={"schema_relation": schema_relation})
        tables = agate_table_get_column_values(table, "memname")
        types = agate_table_get_column_values(table, "type")
        return [
            SasRelation.create(database=DEFAULT_DB, schema=schema_relation.libname, identifier=identifier, type=RelationType(type_))
            for (identifier, type_) in zip(tables, types)
        ]

    def run_sql_for_tests(self, sql, fetch: str, conn):
        """This method is used by the test suite"""
        _, result = self.execute(sql, fetch=fetch in ("one", "all"))
        if fetch == "one":
            return result[0]
        elif fetch == "all":
            return list(result)
        else:
            return

    @classmethod
    def convert_text_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return TEXT

    @classmethod
    def convert_number_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        decimals = agate_table.aggregate(agate.MaxPrecision(col_idx))  # type: ignore[attr-defined]
        return "float" if decimals else "integer"

    @classmethod
    def convert_boolean_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "int"

    @classmethod
    def convert_datetime_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "date"  # TODO

    @classmethod
    def convert_date_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "date"  # TODO

    @classmethod
    def convert_time_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "date"  # TODO

    @available
    def should_identifier_be_quoted(self, identifier, models_column_dict=None) -> bool:
        """Returns True if identifier should be quoted else False"""
        return False

    @available
    def check_and_quote_identifier(self, identifier, models_column_dict=None) -> str:
        return identifier

    @classmethod
    def quote(cls, identifier: str) -> str:
        return identifier

    def create_schema(self, relation: SasRelation) -> None:
        """Create a new library in the lib_base_path directory"""
        sas_log.note(f"Create schema {relation}")
        connection = self.connections.get_thread_connection()
        lib_base_path = connection.credentials.lib_base_path
        if not lib_base_path:
            raise RuntimeException(
                "`lib_base_path` option is required for schema creation - please add the option to the profiles.yml configuration"
            )
        path = path_join(lib_base_path, relation.libname.lower())
        connection.handle.submit(sas_macros.CREATE_SCHEMA.format(libname=relation.libname, path=path))

    def drop_schema(self, relation: SasRelation) -> None:
        sas_log.note(f"Drop schema {relation}")
        connection = self.connections.get_thread_connection()
        lib_base_path = connection.credentials.lib_base_path
        if not lib_base_path:
            raise RuntimeException(
                "`lib_base_path` option is required for dropping schemas - please add the option to the profiles.yml configuration"
            )
        path = path_join(lib_base_path, relation.libname.lower())
        connection.handle.submit(sas_macros.DROP_SCHEMA.format(libname=relation.libname, path=path), ignore_warnings=True)
        self.cache.drop_schema(relation.database, relation.schema)

    def truncate_relation(self, relation: SasRelation) -> None:
        sas_log.note(f"Truncate relation {relation}")
        connection = self.connections.get_thread_connection()
        connection.handle.submit(
            sas_macros.TRUNCATE_DATA_SET.format(
                libname=relation.libname,
                dataset=relation.dataset,
            )
        )

    def rename_relation(self, from_relation: SasRelation, to_relation: SasRelation) -> None:
        sas_log.note(f"Rename relation {from_relation} -> {to_relation}")
        connection = self.connections.get_thread_connection()
        connection.handle.submit(
            sas_macros.RENAME_DATA_SET.format(
                libname=from_relation.libname, old_name=from_relation.dataset, new_name=to_relation.dataset
            )
        )

    def drop_relation(self, relation: SasRelation):
        sas_log.note(f"Drop relation {relation}")
        connection = self.connections.get_thread_connection()
        connection.handle.submit(
            sas_macros.DELETE_DATA_SET.format(
                libname=relation.libname,
                dataset=relation.dataset,
            )
        )

    def get_columns_in_relation(self, relation: SasRelation) -> List[SasColumn]:
        """Get a list of the columns in the given Relation."""
        sas_log.note(f"Get columns in relation {relation}")
        table = self.execute_macro("sas__get_columns_in_relation", kwargs={"relation": relation})
        names = agate_table_get_column_values(table, "name")
        types = agate_table_get_column_values(table, "type")
        lengths = agate_table_get_column_values(table, "length")
        return [
            SasColumn(name, type_, length) for (name, type_, length) in zip(names, types, lengths)
        ]  # TODO check data types (data types are 'num' and 'char')

    def get_relation(self, database: str, schema: str, identifier: str) -> Optional[BaseRelation]:
        return super().get_relation(DEFAULT_DB, schema, identifier)  # Force default db

    def expand_column_types(self, goal: SasRelation, current: SasRelation) -> None:  # type: ignore[override]
        pass

    def expand_target_column_types(self, from_relation: SasRelation, to_relation: SasRelation) -> None:
        pass

    @available
    def import_seed(
        self,
        agate_table: agate.Table,
        relation: SasRelation,
    ) -> Tuple[Any, Any]:
        """Import seed into the SAS server

        1) Store data from an agate table into a CSV file
        2) Upload the CSV file to the SAS server
        3) Import the data into target table (relation)
        4) Delete the local/remote temp files
        """
        connection = self.connections.get_thread_connection()
        local_filename = get_temp_local_filename(ext=".csv")
        remote_filename = get_temp_filename(ext=".csv")
        try:
            # Save the data in a local CSV file
            agate_table.to_csv(local_filename, delimiter=",")  # TODO: check quoted
            # Upload the CSV to the SAS server
            connection.handle.upload_file(local_filename, remote_filename)
            # Import the CSV into the target table
            sas_log.note(f"Import CSV from {remote_filename} into {relation.libname}.{relation.dataset}")
            code = sas_macros.IMPORT_CSV.format(
                libname=relation.libname,
                dataset=relation.dataset,
                filename=remote_filename,
            )
            connection.handle.submit(code)
            sas_log.note("CSV Import done")
            return code
        finally:
            # Delete the local file
            try:
                local_filename.unlink()
            except Exception:
                pass
            # Delete the file from the SAS server
            try:
                sas_log.note(f"Delete file - Filename={remote_filename}")
                code = sas_macros.DELETE_FILE.format(remote_filename=remote_filename)
                connection.handle.submit(code)
            except Exception:
                pass

    # This method only really exists for test reasons.
    def get_rows_different_sql(
        self,
        relation_a: SasRelation,
        relation_b: SasRelation,
        column_names: Optional[List[str]] = None,
        except_operator: str = "EXCEPT",
    ) -> str:
        """Generate SQL for a query that returns a single row with a two
        columns: the number of rows that are different between the two
        relations and the number of mismatched rows.
        """
        names: List[str]
        if column_names is None:
            columns = self.get_columns_in_relation(relation_a)
            names = sorted((c.name for c in columns))
        else:
            names = sorted((n for n in column_names))
        columns_csv = ", ".join(names)
        return sas_macros.COLUMNS_EQUAL.format(
            columns=columns_csv,
            relation_a=relation_a.render(),
            relation_b=relation_b.render(),
            except_op=except_operator,
        )

    @available
    def verify_database(self, database: str) -> str:
        # return an empty string on success so macros can call this
        return ""

    def get_catalog(self, manifest: Manifest) -> Tuple[agate.Table, List[Exception]]:
        for node in itertools.chain(manifest.nodes.values(), manifest.sources.values()):
            node.database = DEFAULT_DB  # Force default db
        return super().get_catalog(manifest)



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import struct
import zlib
from dataclasses import dataclass
from typing import Dict, List, Union

import sqlparse
from sqlparse.sql import (
    Comment,
    Comparison,
    Function,
    Identifier,
    IdentifierList,
    Parenthesis,
    Statement,
    Where,
)
from sqlparse.tokens import Token

from dbt.adapters.sas.constants import DUMMY_FROM, WORK_SCHEMA

__all__ = ["prepare_query", "Query"]

# Remove FINAL from the keyword list
try:
    del sqlparse.keywords.KEYWORDS['FINAL']
except:
    pass


@dataclass
class Query:
    pre: str
    post: str
    query: str
    ctes: Dict[str, str]  # CTEs  identifier => view name


def get_identifier(token: Token) -> str:
    if isinstance(token, Function):
        token = token[0]
    return str(token)


def is_whitespace(token: Token) -> bool:
    return token.ttype in [Token.Text.Whitespace, Token.Text.Whitespace.Newline]


def get_token_skip_whitespace(tokens: List["Token"]) -> Token:
    "Get the next whitespace token from tokens"
    return pop_token_skip_whitespace(list(tokens))


def pop_token_skip_whitespace(tokens: List["Token"]) -> Token:
    "Pop the next whitespace token from tokens (remote the element from tokens)"
    while tokens:
        token = tokens.pop(0)
        if not is_whitespace(token):
            return token
    return None


def process_cte_definition(token: Identifier, pre: List[str], post: List[str], ctes: Dict[str, str]) -> None:
    # print('>>>', token, type(token), token.ttype)
    tokens = list(token.tokens)
    # Get identifier
    token = pop_token_skip_whitespace(tokens)
    identifier = get_identifier(token)
    # identifier = f"{WORK_SCHEMA}.{identifier}"
    # Check 'AS' keyword
    token = pop_token_skip_whitespace(tokens)
    if token.ttype != Token.Keyword or token.value.upper() != "AS":
        raise Exception(f"Unexpected token {token} (`AS` expected)")
    # Get definition
    token = pop_token_skip_whitespace(tokens)
    if not isinstance(token, Parenthesis):
        raise Exception(f"Unexpected token {token} (parenthesis expected)")
    cte_query = []
    process_parenthesis(token, pre, post, cte_query, ctes)
    query = "".join(cte_query)

    crc = zlib.crc32(query.encode('utf-8'))
    encoded_crc = base64.b32encode(struct.pack('I', crc)).decode('ascii').strip('=')[:7]
    view_identifier = f"{WORK_SCHEMA}.{identifier}_{encoded_crc}"
    ctes[identifier.lower()] = view_identifier

    pre.append(f"create view {view_identifier} as {query};")
    post.append(f"drop view {view_identifier};")


def process_cte(statement: Statement, pre: List[str], post: List[str], ctes: Dict[str, str]) -> None:
    token = pop_token_skip_whitespace(statement)
    if isinstance(token, IdentifierList):  # Multiple CTE
        with_tokens = list(token.tokens)
        while with_tokens:
            with_token = pop_token_skip_whitespace(with_tokens)
            if with_token.ttype != Token.Punctuation:
                process_cte_definition(with_token, pre, post, ctes)
    elif isinstance(token, Identifier):  # Single CTE
        process_cte_definition(token, pre, post, ctes)
    else:
        raise Exception(f"Unexpected token {token}")


def process_parenthesis(
    parenthesis: Parenthesis,
    pre: List[str],
    post: List[str],
    query: List[str],
    ctes: Dict[str, str],
    insert_schema: bool = False,
) -> None:
    p_tokens = parenthesis.tokens
    assert p_tokens[0].value == "("
    assert p_tokens[-1].value == ")"
    query.append("(")
    process_tokens(p_tokens[1:-1], pre, post, query, ctes, insert_schema=insert_schema)
    query.append(")")


def process_identifier(
    identifier: Identifier, pre: List[str], post: List[str], query: List[str], ctes: Dict[str, str], insert_schema: bool = False
) -> None:
    i_tokens = identifier.tokens
    has_schema = False
    while i_tokens:
        token = i_tokens.pop(0)
        if isinstance(token, Parenthesis):
            process_parenthesis(token, pre, post, query, ctes)
            has_schema = False
        elif token.ttype == Token.Punctuation and token.value == ",":
            has_schema = False
            query.append(str(token))
        elif token.ttype == Token.Punctuation and token.value == ".":
            has_schema = True
            query.append(str(token))
        elif isinstance(token, Function) and token.tokens[0].value.upper() == "CAST":
            cast_tokens = list(token.tokens[1:])
            token = pop_token_skip_whitespace(cast_tokens)
            if not isinstance(token, Parenthesis):
                raise Exception(f"Unexpected token {token} (`(` expected)")
            parenthesis_tokens = list(token.tokens[1:])
            token = pop_token_skip_whitespace(parenthesis_tokens)
            query.append(str(token.tokens[0]))
        elif isinstance(token, Comment):
            pass  # skip comments
        else:
            identifier = str(token)
            # print('>>>', token, type(token), token.ttype)
            if insert_schema and identifier.lower() in ctes and not has_schema:
                next_token = get_token_skip_whitespace(i_tokens)
                if next_token and next_token.value.upper() == "AS":
                    identifier = ctes[identifier.lower()]  # don't add the alias (already present)
                elif next_token and isinstance(next_token, Identifier):
                    identifier = ctes[identifier.lower()]  # don't add the alias (already present)
                else:
                    identifier = ctes[identifier.lower()] + " as " + identifier.lower()
            query.append(identifier)
            has_schema = False


def process_from(statement: Statement, pre: List[str], post: List[str], query: List[str], ctes: Dict[str, str]) -> None:
    while statement:
        token = statement.pop(0)
        if isinstance(token, Identifier):
            process_identifier(token, pre, post, query, ctes, insert_schema=True)
        elif isinstance(token, IdentifierList):
            id_tokens = list(token.tokens)
            while id_tokens:
                id_token = id_tokens.pop(0)
                if id_token.ttype == Token.Punctuation:
                    query.append(str(id_token))
                elif isinstance(token, Identifier):
                    process_identifier(id_token, pre, post, query, ctes, insert_schema=True)
                else:
                    query.append(str(id_token))
        elif isinstance(token, Parenthesis):
            process_parenthesis(token, pre, post, query, ctes, insert_schema=True)
        elif isinstance(token, Where):
            process_where(token, pre, post, query, ctes, insert_schema=True)
        elif isinstance(token, Comment):
            pass  # skip comments
        else:
            query.append(str(token))


def process_where(
    where: Union[Where, Comparison, Parenthesis],
    pre: List[str],
    post: List[str],
    query: List[str],
    ctes: Dict[str, str],
    insert_schema: bool = False,
) -> None:
    w_tokens = list(where.tokens)
    while w_tokens:
        token = w_tokens.pop(0)
        # if token.ttype != Token.Text.Whitespace and token.ttype != Token.Text.Whitespace.Newline:
        #     print('>>>', token, type(token), token.ttype)
        if token.ttype == Token.Operator.Comparison and token.value == "!=":
            query.append("<>")
        elif token.ttype == Token.Keyword and token.value.upper() == "TRUE":
            query.append("1")
        elif token.ttype == Token.Keyword and token.value.upper() == "FALSE":
            query.append("0")
        elif isinstance(token, Comparison):
            process_where(token, pre, post, query, ctes, insert_schema)
        elif isinstance(token, Parenthesis):
            process_where(token, pre, post, query, ctes, insert_schema)
        elif isinstance(token, Function):
            process_where(token, pre, post, query, ctes, insert_schema)
        elif isinstance(token, Comment):
            pass  # skip comments
        else:
            query.append(str(token))


def process_tokens(
    statement: Statement, pre: List[str], post: List[str], query: List[str], ctes: Dict[str, str], insert_schema=False
) -> None:
    is_a_select: bool = False
    has_from: bool = False
    while statement:
        token = statement.pop(0)
        # if token.ttype != Token.Text.Whitespace and token.ttype != Token.Text.Whitespace.Newline:
        #     print('>>>', token, type(token), token.ttype)
        if token.ttype == Token.Keyword.DML and str(token).upper() == "SELECT":
            is_a_select = True
            query.append(str(token))
        elif token.ttype == Token.Keyword.CTE:
            process_cte(statement, pre, post, ctes)
        elif token.ttype == Token.Keyword and token.value.upper() == "TRUE":
            query.append("1")
        elif token.ttype == Token.Keyword and token.value.upper() == "FALSE":
            query.append("0")
        elif token.ttype == Token.Keyword and token.value.upper().split(" ")[0] == "UNION":
            if is_a_select and not has_from:
                query.append(DUMMY_FROM)
                has_from = False
            query.append(str(token))
        elif isinstance(token, Identifier):
            process_identifier(token, pre, post, query, ctes, insert_schema=insert_schema)
        elif token.ttype == Token.Keyword and token.value.upper() == "FROM":
            query.append(str(token))
            query.append(" ")
            has_from = True
            process_from(statement, pre, post, query, ctes)
        elif isinstance(token, IdentifierList):
            process_tokens(list(token.tokens), pre, post, query, ctes, insert_schema=insert_schema)
        elif isinstance(token, Comment):
            pass  # skip comments
        else:
            query.append(str(token))
    if is_a_select and not has_from:
        query.append(DUMMY_FROM)


def prepare_query(sql: str) -> Query:
    pre: List[str] = []
    post: List[str] = []
    query: List[str] = []
    ctes: Dict[str, str] = {}  # identifier => view name
    for statement in sqlparse.parse(sql):
        statement = list(statement.tokens)
        process_tokens(statement, pre, post, query, ctes)
    # If the query is a "CREATE VIEW" using a CTEs, don't delete the CTEs
    keep_ctes = pre and "".join(query).upper().strip().startswith("CREATE VIEW")
    if keep_ctes:
        post = []
    return Query(
        pre="\n".join(pre),
        post="\n".join(post),
        query="".join(query),
        ctes=ctes,
    )



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = "get_handler"

from typing import Type

from dbt.exceptions import RuntimeException

from .abstract_handler import AbstractConnectionHandler


def get_handler(handler: str) -> Type[AbstractConnectionHandler]:
    if handler == "ws":
        from .ws_handler import WsConnectionHandler

        return WsConnectionHandler
    elif handler == "saspy":
        from .saspy_handler import SaspyConnectionHandler

        return SaspyConnectionHandler
    else:
        raise RuntimeException(f"Invalid handle f{handler}")



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import tempfile
import uuid
from pathlib import Path
from typing import Any, List

import agate

from dbt.adapters.sas.constants import WORK_SCHEMA
from dbt.exceptions import RuntimeException

__all__ = [
    "get_temp_filename",
    "get_temp_local_filename",
    "get_temp_data_set_name",
    "path_join",
    "agate_table_get_column_values",
]


def get_temp_filename(ext: str = "") -> str:
    """Create a temporary filename"""
    return f"/tmp/{str(uuid.uuid4())}{ext}"


def get_temp_local_filename(ext: str = "") -> Path:
    """Create a temporary local filename"""
    return Path(tempfile.gettempdir()) / f"{str(uuid.uuid4())}{ext}"


def get_temp_data_set_name(schema: str = WORK_SCHEMA) -> str:
    """Create a temporary data set name"""
    # The data set name can be up to 32 bytes long for the Base SAS engine starting in Version 7
    uuid_bytes = uuid.uuid4().bytes
    table_name = base64.b32encode(uuid_bytes).decode("ascii").replace("=", "")[:20]
    return f"{schema}.TMP_{table_name}"


def path_join(*parts: List[str]) -> str:
    """Join two or more pathname components"""
    parts = [x for x in parts if x]
    if not parts:
        return ""
    if any([x for x in parts if ".." in x]):
        raise RuntimeException("Invalid path")
    # Count posix separators
    p1 = len([x for x in parts if x.startswith("/") or x.endswith("/")])
    p2 = len([x for x in parts if "/" in x])
    # Count windows separators
    w1 = len([x for x in parts if x.endswith("\\")])
    w2 = len([x for x in parts if "\\" in x])
    if w1 > p1 or w2 > p2:
        separator = "\\"  # Windows separator
    else:
        separator = "/"  # Posix sepatator
    result = separator.join([x.strip(separator) for x in parts if x])
    if separator == "/" and not result.startswith("/") and parts[0].startswith("/"):
        result = "/" + result
    return result


def agate_table_get_column_values(table: agate.Table, column: str) -> List[Any]:
    """Retun the values from an agate.Table's column"""
    col = table.columns.get(column)
    if col is None:
        return []
    else:
        return col.values()



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import struct
import zlib
from dataclasses import dataclass
from typing import Optional, TypeVar

from dbt.adapters.base.relation import BaseRelation, Policy
from dbt.adapters.sas.connections import is_lib_name_strict_mode
from dbt.adapters.sas.constants import DATA_SET_MAX_LEN, DEFAULT_DB, LIBNAME_MAX_LEN
from dbt.contracts.relation import RelationType
from dbt.exceptions import DatabaseException

__all__ = [
    "SasRelation",
    "is_valid_libname",
    "get_short_libname",
]

Self = TypeVar("Self", bound="SasRelation")


@dataclass
class SasQuotePolicy(Policy):
    database: bool = False
    schema: bool = False
    identifier: bool = False


@dataclass
class SasIncludePolicy(Policy):
    database: bool = False  # exclude "database" part from idenfifier
    schema: bool = True
    identifier: bool = True


@dataclass(frozen=True, eq=False, repr=False)
class SasRelation(BaseRelation):
    quote_policy: SasQuotePolicy = SasQuotePolicy()
    include_policy: SasIncludePolicy = SasIncludePolicy()
    quote_character: str = ""

    @property
    def libname(self):
        return self.schema.upper()

    @property
    def dataset(self):
        return self.identifier.upper()

    def include(
        self,
        database: Optional[bool] = None,
        schema: Optional[bool] = None,
        identifier: Optional[bool] = None,
    ) -> Self:
        return super().include(database=False, schema=schema, identifier=identifier)

    @classmethod
    def create(
        cls,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        identifier: Optional[str] = None,
        type: Optional[RelationType] = None,
        **kwargs,
    ) -> Self:
        if is_lib_name_strict_mode():
            if schema:
                schema = schema.lower()
                for prefix in ("dbt_test_dbt_test_", "dbt_test_"):
                    if schema.startswith(prefix):
                        schema = schema[len(prefix) :].strip('_')
                        schema = f"t_{schema}"
                if len(schema) > LIBNAME_MAX_LEN:
                    raise DatabaseException(f"`{schema}` is not a valid libname (length > {LIBNAME_MAX_LEN})")
            if identifier and len(identifier) > DATA_SET_MAX_LEN:
                raise DatabaseException(f"`{identifier}` is not a valid data set name (length > {DATA_SET_MAX_LEN}")
        else:
            if schema:
                schema = schema.lower()
                if not is_valid_libname(schema):
                    schema = get_short_libname(schema)
            if identifier:
                identifier = identifier[:DATA_SET_MAX_LEN]
        return super().create(database=DEFAULT_DB, schema=schema, identifier=identifier, type=type, **kwargs)

    @staticmethod
    def add_ephemeral_prefix(name: str) -> str:
        return f"_cte_{name}"


def is_valid_libname(libname: str) -> bool:
    if libname.lower() == "dictionary":
        return True
    else:
        return len(libname) <= LIBNAME_MAX_LEN


def get_short_libname(libname: str) -> str:
    "Compute the short (8 chars) libname from a long libname"
    if is_valid_libname(libname):
        return libname.lower()
    else:
        crc = zlib.crc32(libname.lower().encode('ascii'))
        return "_" + base64.b32encode(struct.pack('I', crc)).decode('ascii').strip('=')[:7]



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from typing import ContextManager, Tuple

import agate

from dbt.adapters.base import BaseConnectionManager
from dbt.adapters.sas import sas_log, sas_macros
from dbt.adapters.sas.credentials import SasCredentials
from dbt.adapters.sas.cte import prepare_query
from dbt.adapters.sas.handlers import get_handler
from dbt.contracts.connection import AdapterResponse, Connection, ConnectionState
from dbt.exceptions import FailedToConnectException, RuntimeException
from dbt.logger import GLOBAL_LOGGER as logger

__all__ = [
    "SasConnectionManager",
    "is_lib_name_strict_mode",
]


default_credentials: SasCredentials = SasCredentials()


def is_lib_name_strict_mode() -> bool:
    "Lib name strict mode"
    return default_credentials.lib_name_strict_mode


class SasConnectionManager(BaseConnectionManager):
    TYPE = "sas"

    @contextmanager
    def exception_handler(self, sql: str) -> ContextManager:
        try:
            yield
        except Exception as ex:
            logger.debug(f"Error while running:\n{sql}")
            logger.debug(ex)
            raise RuntimeException(str(ex))

    @classmethod
    def open(cls, connection: Connection) -> Connection:
        """Connect to SAS"""
        global default_credentials
        if connection.state == ConnectionState.OPEN:
            logger.debug("Connection is already open, skipping open.")
            return connection
        else:
            try:
                # Open Connection
                handle_class = get_handler(connection.credentials.handler)
                connection.handle = handle_class(connection.credentials)
                connection.state = ConnectionState.OPEN
                default_credentials = connection.credentials
                return connection
            except FailedToConnectException:
                connection.handle = None
                connection.state = ConnectionState.FAIL
                raise

    @classmethod
    def close(cls, connection: Connection) -> Connection:
        """Close connection."""
        if connection.state in {ConnectionState.CLOSED, ConnectionState.INIT}:
            return connection
        else:
            try:
                connection.state = ConnectionState.CLOSED
                connection.handle.endsas()
                connection.handle = None
            except Exception:
                pass
            finally:
                return connection

    def begin(self) -> None:
        """Begin a transaction."""
        pass

    def commit(self) -> None:
        """Commit a transaction."""
        pass

    def clear_transaction(self) -> None:
        """Clear any existing transactions."""
        pass

    def cancel_open(self) -> None:
        """Cancel all open connections on the adapter."""
        pass

    @classmethod
    def get_response(cls, message: str) -> AdapterResponse:
        return AdapterResponse(_message=message)

    def execute(self, sql: str, auto_begin: bool = False, fetch: bool = False) -> Tuple[AdapterResponse, agate.Table]:
        """Execute the given SQL.

        :param str sql: The SQL to execute.
        :param bool auto_begin: Ignored.
        :param bool fetch: If set, fetch results.
        :return: A tuple of the query status and results (empty if fetch=False).
        :rtype: Tuple[AdapterResponse, agate.Table]
        """
        sas_log.note(f"Execute SQL, fetch: {fetch}")
        connection = self.get_thread_connection()
        if fetch and not sql.strip().lower().startswith("select"):
            sas_log.note("Not a select, set fetch to False")
            fetch = False
        if fetch:
            table = self.select(connection, sql)
            message = "SELECT"
        else:
            table = self.proc_sql(connection, sql)
            message = "OK"
        return (self.get_response(message), table)

    def proc_sql(self, connection: Connection, sql: str) -> agate.Table:
        """Execute and SQL Procedure - it does not fetch the result and returns and empty table"""
        sas_log.debug(f"Original SQL: {sql}", family="original_sql")
        query = prepare_query(sql)
        # Create temp tables
        if query.pre:
            sas_log.debug(f"Pre SQL statement: {query.pre}", family="sql")
            code = sas_macros.PROC_SQL.format(sql=query.pre)
            connection.handle.submit(code)
        # Query
        sas_log.debug(f"SQL statement: {query.query}", family="sql")
        code = sas_macros.PROC_SQL.format(sql=query.query)
        connection.handle.submit(code)
        # Drop temp tables
        if query.post:
            sas_log.debug(f"Post SQL statements: {query.post}", family="sql")
            code = sas_macros.PROC_SQL.format(sql=query.post)
            connection.handle.submit(code)
        return agate.Table(rows=[])

    def select(self, connection: Connection, sql: str) -> agate.Table:
        """Execute a SQL select on the SAS server"""
        sas_log.debug(f"Original SQL: {sql}", family="original_sql")
        query = prepare_query(sql)
        # Create temp tables
        if query.pre:
            sas_log.debug(f"Pre SQL statement: {query.pre}", family="sql")
            code = sas_macros.PROC_SQL.format(sql=query.pre)
            connection.handle.submit(code)
        # Query
        sas_log.debug(f"SQL statement: {query.query}", family="sql")
        result = connection.handle.select(query.query)
        # Drop temp tables
        if query.post:
            sas_log.debug(f"Post SQL statements: {query.post}", family="sql")
            code = sas_macros.PROC_SQL.format(sql=query.post)
            connection.handle.submit(code)
        return result

    def delete_file(self, connection: Connection, remote_filename: str) -> None:
        """Delete a file from the SAS server"""
        connection.handle.delete_file(remote_filename)



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Optional, Tuple

from dbt.adapters.base import Credentials
from dbt.adapters.sas.constants import (
    DEFAULT_DB,
    DEFAULT_HANDLER,
    DEFAULT_PORT,
    DEFAULT_WS_PASSWORD,
    DEFAULT_WS_URL,
    DEFAULT_WS_USER,
)

__all__ = ["SasCredentials"]


@dataclass(init=False)
class SasCredentials(Credentials):
    """Connect to SAS via Java IOM"""

    # IOM user
    username: str
    # Password for the IOM user
    password: str
    # Libname
    schema: str
    # Host name, or ip to the IOM server to connect to
    host: str
    # Port IOM is listening on
    port: int = DEFAULT_PORT
    # Path to the java executable
    java_path: Optional[str] = None
    # Timeout value for establishing connection to workspace server
    timeout: Optional[int] = None
    # Autoexec
    autoexec: Optional[str] = None
    # Libraries base path
    lib_base_path: Optional[str] = None
    # Strict mode
    lib_name_strict_mode: bool = False
    # Raise and error if it encounters a warning
    fail_on_warnings: bool = True
    # Ignored
    database: str = DEFAULT_DB
    # Handler
    handler: str = DEFAULT_HANDLER
    # Web service handler username
    ws_basic_user: str = DEFAULT_WS_USER
    # Web service handler password
    ws_basic_password: str = DEFAULT_WS_PASSWORD
    # Web service handler url
    ws_base_url: str = DEFAULT_WS_URL

    _ALIASES = {
        "pass": "password",
        "user": "username",
        "lib": "schema",
    }

    def __init__(self, **kwargs) -> None:
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.database = DEFAULT_DB

    @property
    def type(self) -> str:
        """Return name of adapter."""
        return "sas"

    @property
    def unique_field(self) -> str:
        """
        Hashed and included in anonymous telemetry to track adapter adoption.
        Pick a field that can uniquely identify one team/organization building with this adapter
        """
        return self.host

    def _connection_keys(self) -> Tuple[str]:
        """
        List of keys to display in the `dbt debug` output.
        """
        return ("host", "port", "username", "schema")



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from pathlib import Path
from typing import Optional, Set, Tuple

__all__ = [
    "note",
    "debug",
    "info",
    "warn",
    "warning",
    "error",
    "critical",
    "fatal",
    "enabled",
]


def prepare_logger() -> Tuple[logging.Logger, bool, Set[str]]:
    logger = logging.getLogger("sas-dbt")
    dbt_sas_log = os.environ.get("DBT_SAS_LOG")
    families = set([x.lower() for x in os.environ.get("DBT_SAS_FAMILY", "sql").split(",")])
    if not dbt_sas_log:
        logger.setLevel(logging.ERROR)
        enabled = False
    else:
        log_file = Path(dbt_sas_log)
        if log_file.exists():
            log_file.unlink()
        logger.setLevel(logging.DEBUG)
        f_handler = logging.FileHandler(dbt_sas_log)
        f_format = logging.Formatter("%(message)s")
        # f_format = logging.Formatter("%(asctime)s %(message)s")
        f_handler.setFormatter(f_format)
        logger.addHandler(f_handler)
        enabled = True
    return logger, enabled, families


def is_enabled(family: Optional[str]) -> bool:
    return enabled and ((not family) or (family in families))


def note(msg, family: Optional[str] = None, *args, **kargs) -> None:
    if is_enabled(family):
        msg = f"NOTE: {msg}"
        logger.debug(msg, *args, **kargs)


def debug(msg, family: Optional[str] = None, *args, **kargs) -> None:
    if is_enabled(family):
        msg = f"\n{msg}\n"
        logger.debug(msg, *args, **kargs)


def info(msg, family: Optional[str] = None, *args, **kargs) -> None:
    if is_enabled(family):
        msg = f"INFO: {msg}"
        logger.info(msg, *args, **kargs)


def warn(msg, family: Optional[str] = None, *args, **kargs) -> None:
    if is_enabled(family):
        msg = f"WARNING: {msg}"
        logger.warn(msg, *args, **kargs)


def warning(msg, family: Optional[str] = None, *args, **kargs) -> None:
    if is_enabled(family):
        msg = f"WARNING: {msg}"
        logger.warning(msg, *args, **kargs)


def error(msg, family: Optional[str] = None, *args, **kargs) -> None:
    if is_enabled(family):
        msg = f"ERROR: {msg}"
        logger.error(msg, *args, **kargs)


def critical(msg, family: Optional[str] = None, *args, **kargs) -> None:
    if is_enabled(family):
        msg = f"CRITICAL: {msg}"
        logger.critical(msg, *args, **kargs)


def fatal(msg, family: Optional[str] = None, *args, **kargs) -> None:
    if is_enabled(family):
        msg = f"FATAL: {msg}"
        logger.fatal(msg, *args, **kargs)


logger, enabled, families = prepare_logger()



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dbt.adapters.base import AdapterPlugin
from dbt.adapters.sas.column import SasColumn
from dbt.adapters.sas.connections import SasConnectionManager
from dbt.adapters.sas.credentials import SasCredentials
from dbt.adapters.sas.impl import SasAdapter
from dbt.adapters.sas.relation import SasRelation
from dbt.include import sas

__all__ = ["SasConnectionManager", "SasCredentials", "SasAdapter", "SasRelation", "SasColumn", "Plugin", "sas"]

Plugin = AdapterPlugin(adapter=SasAdapter, credentials=SasCredentials, include_path=sas.PACKAGE_PATH)



#!/usr/bin/en
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import time
from io import BytesIO
from typing import Optional

import agate
import saspy
from saspy.sasexceptions import SASIOConnectionError

from dbt.adapters.sas import sas_log, sas_macros
from dbt.adapters.sas.credentials import SasCredentials
from dbt.adapters.sas.utils import get_temp_data_set_name, get_temp_filename
from dbt.adapters.sas.whereis import whereis
from dbt.clients.agate_helper import DEFAULT_TYPE_TESTER
from dbt.exceptions import FailedToConnectException, ParsingException, RuntimeException
from dbt.logger import GLOBAL_LOGGER as logger

from .abstract_handler import AbstractConnectionHandler

__all__ = ["SaspyConnectionHandler"]


class SaspyConnectionHandler(AbstractConnectionHandler):
    def __init__(self, credentials: SasCredentials) -> None:
        super().__init__(credentials)
        # Load autoexec
        autoexec = self.load_autoexec(credentials)
        # Search for java
        java = credentials.java_path or whereis("java")
        if not java:
            raise FailedToConnectException("Java not found in path.")
        # Open SAS session
        try:
            self.session = saspy.SASsession(
                java=str(java),
                iomhost=credentials.host,
                iomport=credentials.port,
                omruser=credentials.username,
                omrpw=credentials.password,
                timeout=credentials.timeout,
                autoexec=autoexec or None,
            )
        except SASIOConnectionError as ex:
            logger.debug(f"Got an error when attempting to create SAS Session: '{ex}'")
            raise FailedToConnectException(str(ex))

    def submit(self, code: str, note: Optional[str] = None, ignore_warnings: bool = False) -> str:
        """Submit code to the SAS server"""
        if note:
            sas_log.note(note)
        sas_log.debug(code, family="sas")
        result = self.session.submit(code, "text")
        self.check_error(result["LOG"], ignore_warnings=ignore_warnings)
        return result["LOG"]

    def select(self, sql: str) -> agate.Table:
        """Execute a SQL select on the SAS server"""
        temp_table = get_temp_data_set_name()
        temp_filename = get_temp_filename()
        sas_log.note(f"Execute select and save output to file - Filename={temp_filename}")
        sas_log.debug(sql, family="sql")
        code = sas_macros.PROC_SQL_SELECT.format(sql=sql, temp_table=temp_table, temp_filename=temp_filename)
        try:
            # Execute the query and save results as JSON
            self.submit(code)
            # Download response ad JSON
            return self.download_json(temp_filename, family="sql")
        finally:
            # Delete the temp file from the SAS server
            self.delete_file(temp_filename)

    def endsas(self) -> None:
        """Terminate the SAS session, shutting down the SAS process"""
        self.session.endsas()

    def download_file(self, filename: str, family: Optional[str] = None) -> BytesIO:
        """This method downloads a file from the SAS servers file system"""
        sas_log.note(f"Download Filename={filename}")
        logn = self.session._io._logcnt()
        logcodeb = f"\nE3969440A681A24088859985{logn}".encode()

        valid = self.session._io._sb.file_info(filename, quiet=True)
        if valid is None:
            raise RuntimeException(f"{filename} not found")
        elif valid == {}:
            raise RuntimeException(f"{filename} is a directory")

        code = f"filename _sp_updn '{filename}' recfm=F encoding=binary lrecl=4096;"
        self.session._io.submit(code, "text")
        self.session._io.stdin[0].send(b"tom says EOL=DNLOAD                          \n")
        self.session._io.stdin[0].send(b"\ntom says EOL=" + logcodeb + b"\n")

        done = False
        data = b""
        bail = False
        output = BytesIO()
        while not done:
            if os.name == "nt":
                try:
                    rc = self.session._io.pid.wait(0)
                    self.session._io.pid = None
                    self.session._io._sb.SASpid = None
                    raise RuntimeException("SAS process has terminated unexpectedly")
                except RuntimeException:
                    raise
                except Exception:
                    pass
            else:
                rc = os.waitpid(self.session._io.pid, os.WNOHANG)
                if rc[1]:
                    self.session._io.pid = None
                    self.session._io._sb.SASpid = None
                    raise RuntimeException("SAS process has terminated unexpectedly")

            if bail:
                if logcodeb in data:
                    done = True
                    break
            try:
                data = self.session._io.stdout[0].recv(4096)
                if logcodeb in data:
                    data = data.rpartition(logcodeb)[0]
                    done = True
                output.write(data)
            except BlockingIOError:
                data = b""

            if not data:
                time.sleep(0.1)
                try:
                    log = self.session._io.stderr[0].recv(4096)
                except BlockingIOError:
                    log = b""
                if logcodeb in log:
                    bail = True

        if sas_log.enabled:
            output.seek(0, 0)  # rewind
            sas_log.debug(output.getvalue().decode("utf-8"), family=family)

        output.seek(0, 0)  # rewind
        return output

    def download_json(self, filename: str, family: Optional[str]) -> agate.Table:
        """Download an parse a JSON file from the SAS servers file system"""
        try:
            file_content = self.download_file(filename, family=family)
            data = json.load(file_content)
            return agate.Table.from_object(data, column_types=DEFAULT_TYPE_TESTER)
        except json.decoder.JSONDecodeError as ex:
            sas_log.error(f"Error parsing JSON response: {str(ex)}")
            raise ParsingException(f"Error parsing JSON response: {str(ex)}")



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pkgutil

__all__ = [
    "NOOP",
    "PROC_SQL",
    "PROC_SQL_SELECT",
    "DELETE_FILE",
    "RENAME_DATA_SET",
    "DELETE_DATA_SET",
    "TRUNCATE_DATA_SET",
    "IMPORT_CSV",
    "ASSIGN_LIBNAMES",
    "CREATE_SCHEMA",
    "DROP_SCHEMA",
    "COLUMNS_EQUAL",
    "UPLOAD_FILE",
]


def load_sas(filename):
    return pkgutil.get_data(__name__, filename).decode("utf-8")


NOOP = load_sas("code/noop.sas")
PROC_SQL = load_sas("code/proc_sql.sas")
PROC_SQL_SELECT = load_sas("code/proc_sql_select.sas")
DELETE_FILE = load_sas("code/delete_file.sas")
RENAME_DATA_SET = load_sas("code/rename_data_set.sas")
DELETE_DATA_SET = load_sas("code/delete_data_set.sas")
TRUNCATE_DATA_SET = load_sas("code/truncate_data_set.sas")
IMPORT_CSV = load_sas("code/import_csv.sas")
ASSIGN_LIBNAMES = load_sas("code/assign_libnames.sas")
CREATE_SCHEMA = load_sas("code/create_schema.sas")
DROP_SCHEMA = load_sas("code/drop_schema.sas")
COLUMNS_EQUAL = load_sas("code/columns_equal.sql")
UPLOAD_FILE = load_sas("code/upload_file.sas")



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import TypeVar

from dbt.adapters.base.column import Column
from dbt.exceptions import RuntimeException

__all__ = ["SasColumn", "TEXT"]

Self = TypeVar("Self", bound="SasColumn")

# SAS Data types:
# CHAR, CHARACTER, DATE, DEC, DECIMAL, DOUBLE, FLOAT, INT, INTEGER,
# NUM, NUMERIC, REAL, SMALLINT, VARCHAR
# https://support.sas.com/documentation/cdl/en/fedsqlref/67364/HTML/default/viewer.htm#n19bf2z7e9p646n0z224cokuj567.htm


TEXT_SIZE = 32767
TEXT = f"VARCHAR({TEXT_SIZE})"
STRING_DATATYPES = {"char", "character", "varchar", "nchar", "nvarchar"}
INTEGER_DATATYPES = {"int", "integer", "smallint", "tinyint"}
FLOAT_DATATYPES = {"dec", "decimal", "double", "float", "num", "numeric", "real", "smallint"}
NUMBER_DATATYPES = set()
NUMBER_DATATYPES.update(INTEGER_DATATYPES)
NUMBER_DATATYPES.update(FLOAT_DATATYPES)


@dataclass(init=False)
class SasColumn(Column):
    TYPE_LABELS = {
        "STRING": TEXT,
        "TIMESTAMP": "DATE",
        "FLOAT": "FLOAT",
        "INTEGER": "INTEGER",
        "BOOLEAN": "SMALLINT",
    }

    STRING_DATATYPES = STRING_DATATYPES
    INTEGER_DATATYPES = INTEGER_DATATYPES
    FLOAT_DATATYPES = FLOAT_DATATYPES
    NUMBER_DATATYPES = NUMBER_DATATYPES

    def string_size(self) -> int:
        if not self.is_string():
            raise RuntimeException("Called string_size() on non-string field!")
        if self.dtype == "text" or self.char_size is None:
            return TEXT_SIZE
        else:
            return int(self.char_size)

    @property
    def data_type(self) -> str:
        if self.is_string():
            return self.string_type(self.string_size())
        elif self.is_numeric():
            return self.numeric_type(self.dtype, self.numeric_precision, self.numeric_scale)
        else:
            return self.dtype

    @classmethod
    def string_type(cls, size: int) -> str:
        return "varchar({})".format(size)

    def is_float(self) -> bool:
        return self.dtype.lower() in self.FLOAT_DATATYPES

    def is_integer(self) -> bool:
        return self.dtype.lower() in self.INTEGER_DATATYPES

    def is_numeric(self) -> bool:
        return self.dtype.lower() in self.NUMBER_DATATYPES

    def is_string(self) -> bool:
        return self.dtype.lower() in self.STRING_DATATYPES



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
from pathlib import Path
from typing import Optional

IS_WINDOWS = platform.system() == "Windows"
PATH_SEPARATOR = ";" if IS_WINDOWS else ":"

DEFAULT_PATH = [
    "/usr/bin",
    "/usr/sbin",
    "/bin",
    "/sbin",
    "/lib",
    "/lib32",
    "/lib64",
    "/usr/local/bin",
    "/usr/local/sbin",
    "/usr/local/lib",
    "/usr/contrib",
    "/usr/hosts",
]

__all__ = ["whereis"]


def whereis(name: str) -> Optional[Path]:
    """Locate the binary for a command"""
    if not name:
        return None
    if IS_WINDOWS:
        name = name + ".exe"
    for path in os.environ.get("PATH", "").split(PATH_SEPARATOR) or DEFAULT_PATH:
        path = Path(path) / name
        if path.exists():
            return path
    return None



#!/usr/bin/env python
#
# Copyright (c) 2022, Alkemy Spa
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import List, Optional

import agate
import requests
from requests.exceptions import RequestException

from dbt.adapters.sas import sas_log
from dbt.adapters.sas.credentials import SasCredentials
from dbt.clients.agate_helper import DEFAULT_TYPE_TESTER
from dbt.exceptions import DatabaseException, FailedToConnectException

from .abstract_handler import AbstractConnectionHandler

__all__ = ["WsConnectionHandler"]

SessionId = str


class WsConnectionHandler(AbstractConnectionHandler):
    """Connect to SAS via Web Service"""

    def __init__(self, credentials: SasCredentials) -> None:
        super().__init__(credentials)

    def submit(self, code: str, note: Optional[str] = None, ignore_warnings: bool = False) -> str:
        """Submit code to the SAS server"""
        return self._submit(code, session_id=self.get_session_id(), note=note, ignore_warnings=ignore_warnings)

    def _submit(self, code: str, session_id: SessionId, note: Optional[str] = None, ignore_warnings: bool = False) -> str:
        if note:
            sas_log.note(note)
        sas_log.debug(code, family="sas")
        params = {"sessionId": session_id}
        data = json.dumps(
            {
                "code": code,
            }
        )
        response = self.request(path="rsubmit/rs", method="POST", params=params, data=data).json()
        self.check_error(response["output"], ignore_warnings=ignore_warnings)
        return response["output"]

    def select(self, sql: str) -> agate.Table:
        """Execute a SQL select on the SAS server"""
        return self._select(sql, session_id=self.get_session_id())

    def _select(self, sql: str, session_id: SessionId) -> agate.Table:
        params = {"sessionId": session_id, "size": 100000}
        sql = sql.rstrip(";")
        data = json.dumps(
            {
                "query": sql,
            }
        )
        response = self.request(path="sql/select", method="POST", params=params, data=data).json()
        # if sas_log.enabled:
        #    sas_log.debug(str(response), family="sql")
        return agate.Table.from_object(response["results"], column_types=DEFAULT_TYPE_TESTER)

    def endsas(self) -> None:
        """Terminate the SAS session, shutting down the SAS process"""
        pass  # TODO

    def url(self, path: str) -> str:
        """Return the full web service url"""
        return f"{self.credentials.ws_base_url}/{path}"

    def request(self, path: str, method: str = "GET", **kargs):
        headers = {"Content-type": "application/json", "Accept": "application/json"}
        try:
            response = requests.request(
                method=method,
                url=self.url(path),
                auth=(self.credentials.ws_basic_user, self.credentials.ws_basic_password),
                headers=headers,
                **kargs,
            )
        except RequestException as ex:
            raise DatabaseException(str(ex))
        if response.status_code == 401:
            raise DatabaseException("Invalid Web Service credentials")
        if response.status_code != 200:
            try:
                message = response.json()["error"]
            except Exception:
                message = response.text
            sas_log.error(message)
            raise DatabaseException(message)
        return response

    def list_sessions(self) -> List[SessionId]:
        """List sessions"""
        return self.request(method="GET", path="session").json().get("sessions", [])

    def open_session(self) -> SessionId:
        """Open a new session and return the session id"""
        path = f"session/user/{self.credentials.username}"
        params = {
            "password": self.credentials.password,
            "host": self.credentials.host,
            "port": self.credentials.port,
        }
        response = self.request(method="GET", path=path, params=params).json()
        if response["status"] != "success":
            raise DatabaseException("Error opening SAS session")
        session_id = response.get("session").get("sessionId")
        # Load autoexec
        autoexec = self.load_autoexec(self.credentials)
        if autoexec:
            self._submit(autoexec, session_id=session_id)
        return session_id

    def get_session_id(self) -> SessionId:
        """Get the session id of an already existing session or open a new session"""
        try:
            sessions = [
                x
                for x in self.list_sessions()
                if x.get("username") == self.credentials.username
                and x.get("host") == self.credentials.host
                and x.get("port") == self.credentials.port
            ]
            if sessions:
                return sessions[0].get("sessionId")
            else:
                return self.open_session()
        except DatabaseException as ex:
            raise FailedToConnectException(ex.msg)
